#!/usr/bin/env python 
# -*- coding: utf-8 -*-
"""
@File        :  __init__.py.py
@Project     :  ptmchat_develop
@Contact     :  yangshansong
@License     :  (C)Copyright 2021-2022, Hisense
@Description :  None

@Modify Time        @Author         @Version        @Description
------------        -------         --------        ------------
2021/12/6 11:24     yangshansong    1.0             None
"""
